import yaml
import logging
from OSMPythonTools import cachingStrategy, overpass
from structs.landmarks import Landmark, LandmarkType
from structs.preferences import Preferences, Preference
import constants

SIGHTSEEING = LandmarkType(landmark_type='sightseeing')
NATURE = LandmarkType(landmark_type='nature')
SHOPPING = LandmarkType(landmark_type='shopping')


class LandmarkManager:
    logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        strategy = cachingStrategy.JSON(cacheDir=constants.OSM_CACHE_DIR)
        self.query_builder = overpass.Overpass()
        with constants.AMENITY_SELECTORS_PATH.open('r') as f:
            self.amenity_selectors = yaml.safe_load(f)
        with constants.LANDMARK_PARAMETERS_PATH.open('r') as f:
            self.parameters = yaml.safe_load(f)
            # max_distance = parameters['city_bbox_side']


    def get_landmark_lists(self, preferences: Preferences, center_coordinates: tuple[float, float]) -> tuple[list[Landmark], list[Landmark]]:
        '''
        Generate a list of landmarks based on the preferences of the user and the center (ie. start) coordinates.
        The list is then used by the pathfinding algorithm to generate a path that goes through the most interesting landmarks.
        :param preferences: the preferences specified by the user
        :param center_coordinates: the coordinates of the starting point
        '''

        L = []

        # List for sightseeing
        if preferences.sightseeing.score != 0:
            score_func = lambda loc, n_tags: int((10 + n_tags * self.parameters['tag_coeff']) * self.parameters['church_coeff'])  
            L1 = self.fetch_landmarks(self.amenity_selectors['sightseeing'], SIGHTSEEING, center_coordinates, self.parameters['city_bbox_side'], score_func)
            self.correct_score(L1, preferences.sightseeing)
            L += L1

        # List for nature
        if preferences.nature.score != 0:
            score_func = lambda loc, n_tags: int((10 + n_tags * self.parameters['tag_coeff']) * self.parameters['park_coeff'])  
            L2 = self.fetch_landmarks(self.amenity_selectors['nature'], NATURE, center_coordinates, self.parameters['city_bbox_side'], score_func)
            self.correct_score(L2, preferences.nature)
            L += L2

        # List for shopping
        if preferences.shopping.score != 0:
            score_func = lambda loc, n_tags: int((10 + n_tags * self.parameters['tag_coeff']))
            L3 = self.fetch_landmarks(self.amenity_selectors['shopping'], SHOPPING, center_coordinates, self.parameters['city_bbox_side'], score_func)
            self.correct_score(L3, preferences.shopping)
            L += L3

        # remove duplicates
        L = list(set(L))
        L_constrained = self.take_most_important(L, self.parameters['N_important'])
        self.logger.info(f'Generated {len(L)} landmarks around {center_coordinates}, and constrained to {len(L_constrained)} most important ones.')
        return L, L_constrained



    # Take the most important landmarks from the list
    def take_most_important(self, landmarks: list[Landmark], n_max: int) -> list[Landmark]:

        landmarks_sorted = sorted(landmarks, key=lambda x: x.attractiveness, reverse=True)
        return landmarks_sorted[:n_max]



    # Correct the score of a list of landmarks by taking into account preference settings
    def correct_score(self, L: list[Landmark], preference: Preference):

        if len(L) == 0 :
            return

        if L[0].type != preference.type :
            raise TypeError(f"LandmarkType {preference.type} does not match the type of Landmark {L[0].name}")

        for elem in L :
            elem.attractiveness = int(elem.attractiveness*preference.score/500)     # arbitrary computation


    # Function to count elements within a certain radius of a location
    def count_elements_within_radius(self, point: Point, radius: int) -> int:
        
        center_coordinates = (point.x, point.y)
        try:
            landmarks = ox.features_from_point(
                center_point = center_coordinates,
                dist = radius,
                tags = {'building': True} # this is a common tag to give an estimation of the number of elements in the area
            )
            return len(landmarks)
        except ox._errors.InsufficientResponseError:
            return 0    




    def fetch_landmarks(
        self,
        amenity_selectors: list[dict],
        landmarktype: LandmarkType,
        center_coordinates: tuple[float, float],
        distance: int,
        score_function: callable
    ) -> list[Landmark]:
        
        landmarks = ox.features_from_point(
            center_point = center_coordinates,
            dist = distance,
            tags = amenity_selectors
        )
        self.logger.info(f'Fetched {len(landmarks)} landmarks around {center_coordinates}.')

        # cleanup the list
        # remove rows where name is None
        landmarks = landmarks[landmarks['name'].notna()]
        # TODO: remove rows that are part of another building

        ret_landmarks = []
        for element, description in landmarks.iterrows():
            osm_type = element[0]
            osm_id = element[1]
            location = description['geometry']
            n_tags = len(description['nodes']) if type(description['nodes']) == list else 1

            if type(location) == Point:
                location = location
            elif type(location) == Polygon or type(location) == MultiPolygon:
                location = location.centroid
            elif type(location) == LineString:
                location = location.interpolate(location.length/2)

            score = score_function(location, n_tags)
            landmark = Landmark(
                name = description['name'],
                type = landmarktype,
                location = (location.x, location.y),
                osm_type = osm_type,
                osm_id = osm_id,
                attractiveness = score,
                must_do = False,
                n_tags = n_tags
            )
            ret_landmarks.append(landmark)

        return ret_landmarks
